@@ -30,6 +30,7 @@
 
 #include <libmm-glib.h>
 #include <glib.h>
+#include <glib-unix.h>
 #include <glib/gi18n.h>
 #include <glib/gstdio.h>
 #include <gio/gunixinputstream.h>
@@ -348,6 +349,14 @@ tear_down (struct wys_data *data)
 }
 
 
+static gboolean
+terminate (gpointer data)
+{
+  g_main_loop_quit (main_loop);
+  return G_SOURCE_REMOVE;
+}
+
+
 static void
 run (const gchar *modem)
 {
@@ -358,6 +367,18 @@ run (const gchar *modem)
 
   main_loop = g_main_loop_new (NULL, FALSE);
 
+  /** Ignore signals which make systemd refrain from restarting us due
+   * to "Restart=on-failure": SIGHUP, and SIGPIPE.
+   * SIGINT would also match this description but is handled nonetheless
+   * as it's very likely this signal originates from a voluntary user
+   * action.
+   */
+  signal (SIGHUP, SIG_IGN);
+  signal (SIGPIPE, SIG_IGN);
+
+  g_unix_signal_add (SIGINT, terminate, NULL);
+  g_unix_signal_add (SIGTERM, terminate, NULL);
+
   printf (APPLICATION_NAME " started with modem `%s'\n",
           modem ? modem : "(none)");
   g_main_loop_run (main_loop);
@@ -395,41 +416,6 @@ check_machine (const gchar *machine)
 }
 
 
-static void
-terminate (int signal)
-{
-  if (main_loop)
-    {
-      g_main_loop_quit (main_loop);
-    }
-}
-
-
-/** Ignore signals which make systemd refrain from restarting us due
- * to "Restart=on-failure": SIGHUP, SIGINT, and SIGPIPE.
- */
-static void
-setup_signals ()
-{
-  void (*ret)(int);
-
-#define try_setup(NUM,handler)                          \
-  ret = signal (SIG##NUM, handler);                     \
-  if (ret == SIG_ERR)                                   \
-    {                                                   \
-      g_error ("Error setting signal handler: %s",      \
-               g_strerror (errno));                     \
-    }
-
-  try_setup (HUP,  SIG_IGN);
-  try_setup (INT,  SIG_IGN);
-  try_setup (PIPE, SIG_IGN);
-  try_setup (TERM, terminate);
-
-#undef try_setup
-}
-
-
 /** This function will close @fd */
 static gchar *
 read_machine_conf_file (const gchar *filename,
@@ -637,8 +623,6 @@ main (int argc, char **argv)
 
   ensure_alsa_card (machine, "WYS_MODEM", "modem", &modem);
 
-  setup_signals ();
-
   run (modem);
 
   return 0;
